@@ -16,7 +16,7 @@ use DynaLoader();
 use Carp ();
 @ISA = qw(DynaLoader);
 
-$VERSION = '2.01_22';
+$VERSION = '2.0200';
 
 bootstrap DBD::mSQL $VERSION;
 
@@ -125,8 +125,6 @@ sub connect {
 	return undef;
     }
 
-    $this->{'private_DBD_mSQL_attr'} = $privateAttrHash;
-
     # Call msqlConnect func in mSQL.xs file
     # and populate internal handle data.
     DBD::mSQL::db::_login($this, $dsn, $username, $password)
@@ -306,8 +304,7 @@ Interface (DBI)
         or
 
     $driver = "mysql";
-    $dsn = "DBI:$driver:database=$database;host=$hostname;port=$port";
-    if ($compression) { $dsn .= ";mysql_compression=1"; }
+    $dsn = "DBI:$driver:database=$database;$options";
 
     $dbh = DBI->connect($dsn, $user, $password);
 
@@ -336,19 +333,6 @@ Interface (DBI)
     $rc = $dbh->func('reload', 'admin');
 
 
-=head1 EXPERIMENTAL SOFTWARE
-
-This package contains experimental software and should *not* be used
-in a production environment. We are following the Linux convention and
-treat the "even" releases (1.18xx as of this writing, perhaps 1.20xx,
-1.22xx, ... in the future) as stable. Only bug or portability fixes
-will go into these releases.
-
-The "odd" releases (1.19xx as of this writing, perhaps 1.21xx, 1.23xx
-in the future) will be used for testing new features or other serious
-code changes.
-
-
 =head1 DESCRIPTION
 
 <DBD::mysql> and <DBD::mSQL> are the Perl5 Database Interface drivers for
@@ -366,23 +350,27 @@ of the I<Msql-Mysql-modules> package.
 
     $driver = "mSQL";  #  or "mSQL1"
     $dsn = "DBI:$driver:$database";
-    $dsn = "DBI:$driver:database=$database;host=$hostname";
-    $dsn = "DBI:$driver:database=$database;host=$hostname;port=$port";
+    $dsn = "DBI:$driver:database=$database;$options";
 
     $dbh = DBI->connect($dsn, undef, undef);
 
         or
 
     $dsn = "DBI:mysql:$database";
-    $dsn = "DBI:mysql:database=$database;host=$hostname";
-    $dsn = "DBI:mysql:database=$database;host=$hostname;port=$port";
-
-    if ($compression) { $dsn .= ";mysql_compression=1"; }
+    $dsn = "DBI:mysql:database=$database;$options";
 
     $dbh = DBI->connect($dsn, $user, $password);
 
 A C<database> must always be specified.
 
+Possible options are, separated by semicolon:
+
+=over 8
+
+=item host
+
+=item port
+
 The hostname, if not specified or specified as '', will default to an
 mysql or mSQL daemon running on the local machine on the default port
 for the UNIX socket.
@@ -390,12 +378,26 @@ for the UNIX socket.
 Should the mysql or mSQL daemon be running on a non-standard port number,
 you may explicitly state the port number to connect to in the C<hostname>
 argument, by concatenating the I<hostname> and I<port number> together
-separated by a colon ( C<:> ) character.
+separated by a colon ( C<:> ) character or by using the  C<port> argument.
+
+=item mysql_compression
 
 As of MySQL 3.22.3, a new feature is supported: If your DSN contains
 the option "mysql_compression=1", then the communication between client
 and server will be compressed.
 
+=item mysql_socket
+
+As of MySQL 3.21.15, it is possible to choose the Unix socket that is
+used for connecting to the server. This is done, for example, with
+
+    mysql_socket=/dev/mysql
+
+Usually there's no need for this option, unless you are using another
+location for the socket than that built into the client.
+
+=back
+
 =back
 
 =head2 Private MetaData Methods
@@ -832,11 +834,22 @@ call. If another connect call uses another port and the handles
 are used simultaneously, they will interfere. I doubt that this
 will ever change.
 
+Msql-2.0.4 and 2.0.4.1 contain a bug that makes ORDER BY and hence
+the test script C<t/40bindparam> fail. To verify, if this is the
+case for you, do a
+
+      cd Msql
+      perl -w -I../blib/lib -I../blib/arch t/40bindparam.t
+
+If something is wrong, the script ought to print a number of id's and
+names. If the id's aren't in order, it is likely, that your mSQL has
+a bug. See the INSTALL file for a patch.
+
 
 =head1 AUTHORS
 
 B<DBD::mSQL> has been primarily written by Alligator Descartes
-(I<descarte@hermetica.com>), who has been aided and abetted by Gary
+(I<descarte@arcana.co.uk>), who has been aided and abetted by Gary
 Shea, Andreas Koenig and Tim Bunce amongst others. Apologies if your
 name isn't listed, it probably is in the file called
 'Acknowledgments'. As of version 0.80 the maintainer is Andreas Knig.
@@ -858,7 +871,7 @@ current version and all credits and copyright notices are retained (
 the I<AUTHOR> and I<COPYRIGHT> sections ).  Requests for other
 distribution rights, including incorporation into commercial products,
 such as books, magazine articles or CD-ROMs should be made to
-Alligator Descartes <I<descarte@hermetica.com>>.
+Alligator Descartes <I<descarte@arcana.co.uk>>.
 
 
 =head1 MAILING LIST SUPPORT
@@ -895,7 +908,7 @@ Mailing list archives are at
 Additional information on the DBI project can be found on the World
 Wide Web at the following URL:
 
-    http://www.hermetica.com/technologia/perl/DBI
+    http://www.arcana.co.uk/technologia/perl/DBI
 
 where documentation, pointers to the mailing lists and mailing list
 archives and pointers to the most current versions of the modules can
