@@ -16,7 +16,7 @@ use vars qw($QUIET @ISA @EXPORT @EXPORT_OK $VERSION $db_errstr);
 $db_errstr = '';
 $QUIET  = 0;
 @ISA    = qw(DBI); # Inherits Exporter and DynaLoader via DBI
-$VERSION = '1.19_22';
+$VERSION = '1.2000';
 
 # @EXPORT is a relict from old times...
 @EXPORT = qw(
@@ -50,12 +50,7 @@ sub FETCH ($$) {
 	$key = $FETCH_map->{$key};
     }
     my($dbh) = $self->{'dbh'};
-    my($attr) = $dbh->{'private_DBD_mysql_attr'};
-    if (exists($attr->{$key})) {
-	$attr->{$key};
-    } else {
-	$dbh->{$key};
-    }
+    $dbh->{$key};
 }
 
 sub STORE ($$$) {
@@ -67,6 +62,16 @@ sub STORE ($$$) {
     }
 }
 
+sub DESTROY {
+    my $self = shift;
+    my $dbh = $self->{'dbh'};
+    if ($dbh) {
+	local $SIG{'__WARN__'} = sub { };
+	$dbh->disconnect();
+    }
+}
+
+
 sub connect ($;$$$$) {
     my($class, $host, $db, $user, $password) = @_;
     my($self) = { 'host' => ($host || ''),
@@ -790,6 +795,8 @@ recommend anybody to carefully watch the development of this module
 be supported for a long time. But it's a dead end. I expect in the
 medium term, that the DBI efforts result in a richer module family
 with better support and more functionality. Alligator maintains an
-interesting page on the DBI development: http://www.hermetica.com/
+interesting page on the DBI development:
+
+    http://www.arcana.co.uk/technologia/perl/DBI
 
 =cut
