@@ -96,6 +96,227 @@ void do_warn(SV* h, int rc, char* what) {
 }
 
 
+
+
+/***************************************************************************
+ *
+ *  Name:    _MyLogin, MyConnect
+ *
+ *  Purpose: Replacements for mysql_connect or msqlConnect
+ *
+ *  Input:   imp_dbh - database handle
+ *
+ *  Returns: TRUE for success, FALSE otherwise; you have to call
+ *           do_error in the latter case.
+ *
+ *  Bugs:    The msql version needs to set the environment
+ *           variable MSQL_TCP_PORT. There's absolutely no
+ *           portable way of setting environment variables
+ *           from within C: Neither setenv() nor putenv()
+ *           are guaranteed to work. I have decided to use
+ *           the internal perl functions setenv_getix()
+ *           and my_setenv() instead, let's hope, this is safe.
+ *
+ *           Another problem was pointed out by Andreas:
+ *           This isn't thread safe. We'll have fun with perl
+ *           5.005 ... :-)
+ *
+ **************************************************************************/
+
+int MyConnect(dbh_t *sock, char* unixSocket, char* host, char* port,
+	      char* user, char* password, char* dbname, imp_dbh_t *imp_dbh) {
+    int portNr;
+
+    if (host && !*host) host = NULL;
+    if (port && *port) {
+        portNr = atoi(port);
+    } else {
+        portNr = 0;
+    }
+    if (user && !*user) user = NULL;
+    if (password && !*password) password = NULL;
+
+    if (dbis->debug >= 2)
+        fprintf(DBILOGFP,
+		"imp_dbh->MyConnect: host = %s, port = %d, uid = %s," \
+                " pwd = %s\n",
+		host ? host : "NULL", portNr,
+		user ? user : "NULL",
+		password ? password : "NULL");
+
+#ifdef DBD_MYSQL
+    {
+#ifndef HAVE_MYSQL_REAL_CONNECT
+        /*
+	 *  Setting a port for mysql's client is ugly: We have to use
+	 *  the not documented variable mysql_port.
+	 */
+        int result, oldPort = 0;
+	if (portNr) {
+	    oldPort = mysql_port;
+	    mysql_port = portNr;
+	}
+        result = mysql_connect(*sock, host, user, password) ? TRUE : FALSE;
+	if (oldPort) {
+	    mysql_port = oldPort;
+	}
+	return result;
+#elif defined(MYSQL_VERSION_ID)  &&  (MYSQL_VERSION_ID >= 32200)
+	mysql_init(*sock);
+#if defined(MYSQL_VERSION_ID)  &&  (MYSQL_VERSION_ID >= 32203)
+	if (imp_dbh  &&  imp_dbh->compression) {
+	    mysql_options(*sock, MYSQL_OPT_COMPRESS, NULL);
+	}
+#endif
+	return mysql_real_connect(*sock, host, user, password, dbname,
+				  portNr, unixSocket, 0) ?
+	    TRUE : FALSE;
+#elif defined(MYSQL_VERSION_ID)  &&  (MYSQL_VERSION_ID >= 32115)
+	return mysql_real_connect(*sock, host, user, password, portNr,
+				  unixSocket, 0) ?
+	    TRUE : FALSE;
+#else
+	return mysql_real_connect(*sock, host, user, password, portNr,
+				  unixSocket) ?
+	    TRUE : FALSE;
+#endif
+    }
+#else
+    {
+        /*
+	 *  Setting a port for msql's client is extremely ugly: We have
+	 *  to set an environment variable. Even worse, we cannot trust
+	 *  in setenv or putenv being present, thus we need to use
+	 *  internal, not documented, perl functions. :-(
+	 */
+        char buffer[32];
+	char* oldPort = NULL;
+
+	sprintf(buffer, "%d", portNr);
+	if (portNr) {
+	    oldPort = environ[setenv_getix("MSQL_TCP_PORT")];
+	    if (oldPort) {
+	        char* copy = (char*) malloc(strlen(oldPort)+1);
+		if (!copy) {
+		    return FALSE;
+		}
+		strcpy(copy, oldPort);
+		oldPort = copy;
+	    }
+	    my_setenv("MSQL_TCP_PORT", buffer);
+	}
+	*sock = msqlConnect(host);
+	if (oldPort) {
+	    my_setenv("MSQL_TCP_PORT", oldPort);
+	    if (oldPort) { free(oldPort); }
+	}
+	return (*sock == -1) ? FALSE : TRUE;
+    }
+#endif
+}
+
+static int _MyLogin(imp_dbh_t *imp_dbh) {
+    SV* sv;
+    SV** svp;
+    HV* hv;
+    char* dbname;
+    char* host;
+    char* port;
+    char* user;
+    char* password;
+    char* unixSocket = NULL;
+    STRLEN len;
+
+    sv = DBIc_IMP_DATA(imp_dbh);
+    if (!sv  ||  !SvROK(sv)) {
+        return FALSE;
+    }
+    hv = (HV*) SvRV(sv);
+    if (SvTYPE(hv) != SVt_PVHV) {
+        return FALSE;
+    }
+    if ((svp = hv_fetch(hv, "host", 4, FALSE))) {
+        host = SvPV(*svp, len);
+	if (!len) {
+	    host = NULL;
+	}
+    } else {
+        host = NULL;
+    }
+    if ((svp = hv_fetch(hv, "port", 4, FALSE))) {
+        port = SvPV(*svp, na);
+    } else {
+        port = NULL;
+    }
+    if ((svp = hv_fetch(hv, "user", 4, FALSE))) {
+        user = SvPV(*svp, len);
+	if (!len) {
+	    user = NULL;
+	}
+    } else {
+        user = NULL;
+    }
+    if ((svp = hv_fetch(hv, "password", 8, FALSE))) {
+        password = SvPV(*svp, len);
+	if (!len) {
+	    password = NULL;
+	}
+    } else {
+        password = NULL;
+    }
+    if ((svp = hv_fetch(hv, "database", 8, FALSE))) {
+        dbname = SvPV(*svp, na);
+    } else {
+        dbname = NULL;
+    }
+#ifdef DBD_MYSQL
+    if ((svp = hv_fetch(hv, "mysql_compression", 17, FALSE))  &&
+	*svp  &&  SvTRUE(*svp)) {
+        imp_dbh->compression = 1;
+    } else {
+        imp_dbh->compression = 0;
+    }
+    if ((svp = hv_fetch(hv, "mysql_socket", 12, FALSE))  &&
+	*svp  &&  SvTRUE(*svp)) {
+        unixSocket = SvPV(*svp, na);
+    }
+#endif
+
+    if (dbis->debug >= 2)
+        fprintf(DBILOGFP,
+		"imp_dbh->MyLogin: dbname = %s, uid = %s, pwd = %s," \
+		"host = %s, port = %s\n",
+		dbname ? dbname : "NULL",
+		user ? user : "NULL",
+		password ? password : "NULL",
+		host ? host : "NULL",
+		port ? port : "NULL");
+
+#ifdef DBD_MYSQL
+    imp_dbh->svsock = &imp_dbh->mysql;
+#endif
+
+#if defined(DBD_MYSQL)  &&  (MYSQL_VERSION_ID >= 32200)
+    if (!MyConnect(&imp_dbh->svsock, unixSocket, host, port, user, password,
+		   dbname, imp_dbh)) {
+        return FALSE;
+    }
+#else
+    if (!MyConnect(&imp_dbh->svsock, unixSocket, host, port, user, password,
+		   NULL, imp_dbh)) {
+        return FALSE;
+    }
+
+    if (MySelectDb(imp_dbh->svsock, dbname)) {
+        MyClose(imp_dbh->svsock);
+	return FALSE;
+    }
+#endif
+
+    return TRUE;
+}
+
+
 /***************************************************************************
  *
  *  Name:    dbd_db_login
@@ -113,11 +334,12 @@ void do_warn(SV* h, int rc, char* what) {
  *           been called in the latter case
  *
  **************************************************************************/
+
 int dbd_db_login(SV* dbh, imp_dbh_t* imp_dbh, char* dbname, char* user,
 		 char* password) {
-    char* copy = NULL;
-    char* host = NULL;
-    char* ptr;
+#ifdef dTHR
+    dTHR;
+#endif
 
     if (dbis->debug >= 2)
         fprintf(DBILOGFP, "imp_dbh->connect: dsn = %s, uid = %s, pwd = %s\n",
@@ -125,50 +347,16 @@ int dbd_db_login(SV* dbh, imp_dbh_t* imp_dbh, char* dbname, char* user,
 	       user ? user : "NULL",
 	       password ? password : "NULL");
 
-    /*
-     *  dbname may be "db:host" or "db;host"
-     */
-    if ((ptr = strchr(dbname, ':'))  ||  (ptr = strchr(dbname, ';'))) {
-        int len = ptr-dbname;
-	copy = (char*) malloc(strlen(dbname)+1);
-	strcpy(copy, dbname);
-	dbname = copy;
-	host = dbname + len;
-	*host++ = '\0';
-    }
-
-    /*
-     *  Try to connect
-     */
-#xtract Mysql
-    imp_dbh->svsock = &imp_dbh->mysql;
-    if (!dbd_db_connect(imp_dbh->svsock, host, user, password, dbname)) {
-#xtract Msql
-    if (!dbd_db_connect(&imp_dbh->svsock, host, user, password, NULL)) {
-#endxtract
-	DO_ERROR(dbh, JW_ERR_CONNECT, imp_dbh->svsock);
-	if (copy) free(copy);
+    if (!_MyLogin(imp_dbh)) {
+	DO_ERROR(dbh, MyErrno(imp_dbh->svsock, JW_ERR_CONNECT),
+		 imp_dbh->svsock);
 	return FALSE;
     }
 
-#if !defined(DBD_MYSQL)  ||  !defined(MYSQL_VERSION_ID)  ||  (MYSQL_VERSION_ID < 32200)
-    /*
-     *  Connected, now try to login
-     */
-    if (MySelectDb(imp_dbh->svsock, dbname)) {
-        if (copy) free(copy);
-	DO_ERROR(dbh, JW_ERR_SELECT_DB, imp_dbh->svsock);
-	MyClose(imp_dbh->svsock);
-	return FALSE;
-    }
-#endif
-
-    if (copy) free(copy);
-
     /*
      *  Tell DBI, that dbh->disconnect should be called for this handle
      */
-    DBIc_on(imp_dbh, DBIcf_ACTIVE);
+    DBIc_ACTIVE_on(imp_dbh);
 
     /*
      *  Tell DBI, that dbh->destroy should be called for this handle
@@ -225,9 +413,13 @@ int dbd_db_rollback(SV* dbh, imp_dbh_t* imp_dbh) {
  **************************************************************************/
 
 int dbd_db_disconnect(SV* dbh, imp_dbh_t* imp_dbh) {
+#ifdef dTHR
+    dTHR;
+#endif
+
     /* We assume that disconnect will always work       */
     /* since most errors imply already disconnected.    */
-    DBIc_off(imp_dbh, DBIcf_ACTIVE);
+    DBIc_ACTIVE_off(imp_dbh);
     if (dbis->debug >= 2)
         fprintf(DBILOGFP, "imp_dbh->svsock: %lx\n", (long) &imp_dbh->svsock);
     MyClose(imp_dbh->svsock );
@@ -253,7 +445,7 @@ int dbd_db_disconnect(SV* dbh, imp_dbh_t* imp_dbh) {
  **************************************************************************/
 
 int dbd_discon_all (SV *drh, imp_drh_t *imp_drh) {
-#if defined (dTHR)
+#if defined(dTHR)
     dTHR;
 #endif
 
@@ -363,30 +555,76 @@ SV* dbd_db_FETCH_attrib(SV* dbh, imp_dbh_t* imp_dbh, SV* keysv) {
     STRLEN kl;
     char *key = SvPV(keysv, kl);
 
-    if (kl==10 && strEQ(key, "AutoCommit")){
-        /*
-	 *  We do support neither transactions nor "AutoCommit".
-	 *  But we stub it. :-)
-	 */
-        return &sv_yes;
-    }
-        
-    if (kl == 5  &&  strEQ(key, "errno")) {
+    switch (*key) {
+      case 'A':
+	if (strEQ(key, "AutoCommit")){
+	    /*
+	     *  We do support neither transactions nor "AutoCommit".
+	     *  But we stub it. :-)
+	     */
+	    return &sv_yes;
+	}
+	break;
+      case 'e':
+	if (strEQ(key, "errno")) {
 #if defined(DBD_MYSQL)  &&  defined(mysql_errno)
-	return sv_2mortal(newSViv((IV)mysql_errno(imp_dbh->svsock)));
+	    return sv_2mortal(newSViv((IV)mysql_errno(imp_dbh->svsock)));
 #else
-	return sv_2mortal(newSViv(-1));
+	    return sv_2mortal(newSViv(-1));
+#endif
+	} else if (strEQ(key, "errmsg")) {
+	    char* msg = MyError(imp_dbh->svsock);
+	    return sv_2mortal(newSVpv(msg, strlen(msg)));
+	}
+	break;
+      case 'h':
+	if (strEQ(key, "hostinfo")) {
+	    char* hostinfo = MyGetHostInfo(imp_dbh->svsock);
+	    return hostinfo ?
+	        sv_2mortal(newSVpv(hostinfo, strlen(hostinfo))) : &sv_undef;
+	}
+	break;
+#if defined(DBD_MYSQL)
+      case 'i':
+	if (strEQ(key, "info")) {
+	    char* info = mysql_info(imp_dbh->svsock);
+	    return info ? sv_2mortal(newSVpv(info, strlen(info))) : &sv_undef;
+	}
+	break;
+#endif
+      case 'p':
+	if (kl == 9  &&  strEQ(key, "protoinfo")) {
+	    return sv_2mortal(newSViv(MyGetProtoInfo(imp_dbh->svsock)));
+	}
+	break;
+      case 's':
+	if (kl == 10  &&  strEQ(key, "serverinfo")) {
+	    char* serverinfo = MyGetServerInfo(imp_dbh->svsock);
+	    return serverinfo ?
+	      sv_2mortal(newSVpv(serverinfo, strlen(serverinfo))) : &sv_undef;
+	} else if (strEQ(key, "sock")) {
+	    return sv_2mortal(newSViv((IV) imp_dbh->svsock));
+#if defined DBD_MYSQL
+	} else if (strEQ(key, "sockfd")) {
+	    return sv_2mortal(newSViv((IV) imp_dbh->svsock->net.fd));
+#endif
+	} else if (strEQ(key, "stats")) {
+#if defined(DBD_MYSQL)
+	    char* stats = mysql_stat(imp_dbh->svsock);
+	    return stats ?
+	        sv_2mortal(newSVpv(stats, strlen(stats))) : &sv_undef;
+#elif defined(DBD_MSQL) && defined(IDX_TYPE)
+	    return sv_2mortal(newSViv(msqlGetServerStats(imp_dbh->svsock)));
+#endif
+	}
+	break;
+#if defined(DBD_MYSQL)
+      case 't':
+	if (kl == 9  &&  strEQ(key, "thread_id")) {
+	    return sv_2mortal(newSViv(mysql_thread_id(imp_dbh->svsock)));
+	}
+	break;
 #endif
-    } else if (kl == 6  &&  strEQ(key, "errmsg")) {
-	char* msg = MyError(imp_dbh->svsock);
-	return sv_2mortal(newSVpv(msg, strlen(msg)));
-#xtract Mysql
-    } else if (kl == 4  &&  strEQ(key, "info")) {
-        char* info = mysql_info(imp_dbh->svsock);
-        return info ? sv_2mortal(newSVpv(info, strlen(info))) : &sv_undef;
-    } else if (kl == 9  &&  strEQ(key, "thread_id")) {
-        return sv_2mortal(newSViv(mysql_thread_id(imp_dbh->svsock)));
-#endxtract
     }
     return Nullsv;
 }
@@ -424,14 +662,14 @@ int dbd_st_prepare(SV* sth, imp_sth_t* imp_sth, char* statement, SV* attribs) {
     imp_sth->done_desc = 0;
     imp_sth->cda = NULL;
     imp_sth->currow = 0;
-#xtract Mysql
+#if defined(DBD_MYSQL)
     imp_sth->use_mysql_use_result = 0;
-#endxtract
+#endif
     for (i = 0;  i < AV_ATTRIB_LAST;  i++) {
 	imp_sth->av_attr[i] = Nullav;
     }
 
-#xtract Mysql
+#if defined(DBD_MYSQL)
     /*
      *  Check attributes
      */
@@ -444,7 +682,7 @@ int dbd_st_prepare(SV* sth, imp_sth_t* imp_sth, char* statement, SV* attribs) {
 	    }
 	}
     }
-#endxtract
+#endif
 
     /*
      *  Allocate memory for parameters
@@ -487,65 +725,121 @@ int dbd_st_internal_execute(SV* h, SV* statement, SV* attribs, int numParams,
 	}
     }
 
-    if (*cdaPtr) { MyFreeResult(*cdaPtr); *cdaPtr = NULL; }
-
     if (slen >= 10
 	&&  tolower(sbuf[0]) == 'l'
 	&&  tolower(sbuf[1]) == 'i'
 	&&  tolower(sbuf[2]) == 's'
-	&&  tolower(sbuf[3]) == 't'
-	&&  tolower(sbuf[4]) == 'f'
-	&&  tolower(sbuf[5]) == 'i'
-	&&  tolower(sbuf[6]) == 'e'
-	&&  tolower(sbuf[7]) == 'l'
-	&&  tolower(sbuf[8]) == 'd'
-	&&  tolower(sbuf[9]) == 's') {
-	char* table;
+	&&  tolower(sbuf[3]) == 't') {
+        if (slen >= 11
+	    &&  tolower(sbuf[4]) == 'f'
+	    &&  tolower(sbuf[5]) == 'i'
+	    &&  tolower(sbuf[6]) == 'e'
+	    &&  tolower(sbuf[7]) == 'l'
+	    &&  tolower(sbuf[8]) == 'd'
+	    &&  tolower(sbuf[9]) == 's'
+	    &&  isspace(sbuf[10])) {
+	    char* table;
 
-	while (slen && !isspace(*sbuf)) { --slen;  ++sbuf; }
-	while (slen && isspace(*sbuf)) { --slen;  ++sbuf; }
+	    slen -= 10;
+	    sbuf += 10;
+	    while (slen && isspace(*sbuf)) { --slen;  ++sbuf; }
 
-	if (!slen) {
-	    do_error(h, JW_ERR_QUERY, "Missing table name");
-	    return -2;
+	    if (!slen) {
+	        do_error(h, JW_ERR_QUERY, "Missing table name");
+		return -2;
+	    }
+
+	    if (!(table = malloc(slen+1))) {
+	        do_error(h, JW_ERR_MEM, "Out of memory");
+		return -2;
+	    }
+	    strncpy(table, sbuf, slen);
+	    sbuf = table;
+	    while (slen && !isspace(*sbuf)) { --slen;  ++sbuf; }
+	    *sbuf++ = '\0';
+
+	    *cdaPtr = MyListFields(svsock, table);
+	    free(table);
+
+	    if (!(*cdaPtr)) {
+	        DO_ERROR(h, JW_ERR_LIST_FIELDS, svsock);
+		return -2;
+	    }
+
+	    return 0;
+#if defined(DBD_MSQL)  &&  defined(IDX_TYPE)
+	} else if (tolower(sbuf[4]) == 'i'
+		   &&  tolower(sbuf[5]) == 'n'
+		   &&  tolower(sbuf[6]) == 'd'
+		   &&  tolower(sbuf[7]) == 'e'
+		   &&  tolower(sbuf[8]) == 'x'
+		   &&  isspace(sbuf[9])) {
+	    char* table;
+	    char* index;
+
+	    slen -= 9;
+	    sbuf += 9;
+
+	    while (slen && isspace(*sbuf)) { --slen;  ++sbuf; }
+	    if (!slen) {
+	        do_error(h, JW_ERR_QUERY, "Missing table name");
+		return -2;
+	    }
+	    if (!(table = malloc(slen+1))) {
+	        do_error(h, JW_ERR_MEM, "Out of memory");
+		return -2;
+	    }
+	    strncpy(table, sbuf, slen);
+	    sbuf = table;
+
+	    while (slen && !isspace(*sbuf)) { --slen;  ++sbuf; }
+	    if (slen) {
+	        *sbuf++ = '\0';
+		--slen;
+	    }
+	    while (slen && isspace(*sbuf)) { --slen;  ++sbuf; }
+	    if (!slen) {
+	        do_error(h, JW_ERR_QUERY, "Missing index name");
+		free(table);
+		return -2;
+	    }
+	    index = sbuf;
+	    while (slen && !isspace(*sbuf)) { --slen;  ++sbuf; }
+	    *sbuf++ = '\0';
+
+	    *cdaPtr = msqlListIndex(svsock, table, index);
+	    free(table);
+	    if (!(*cdaPtr)) {
+	        DO_ERROR(h, JW_ERR_LIST_INDEX, svsock);
+		return -2;
+	    }
+
+	    return 0;
+#endif
 	}
-
-	if (!(table = malloc(slen+1))) {
-	    do_error(h, JW_ERR_MEM, "Out of memory");
-	    return -2;
-	}
-	strncpy(table, sbuf, slen);
-	table[slen] = '\0';
-	*cdaPtr = MyListFields(svsock, sbuf);
-	free(table);
-
-	if (!(*cdaPtr)) {
-	    DO_ERROR(h, JW_ERR_LIST_FIELDS, svsock);
-	    return -2;
-	}
-
-	return 0;
-    } else {
-	if (MyQuery(svsock, sbuf, slen) == -1) {
-	    Safefree(salloc);
-	    DO_ERROR(h, JW_ERR_QUERY, svsock);
-	    return -2;
-	}
-	Safefree(salloc);
-
-	/** Store the result from the Query */
-#xtract Mysql
-	if (!(*cdaPtr = (use_mysql_use_result ?
-		 mysql_use_result(svsock) : mysql_store_result(svsock)))) {
-	    return mysql_affected_rows(svsock);
-#xtract Msql
-	if (!(*cdaPtr = MyStoreResult(svsock))) {
-	    return -1;
-#endxtract
-	}
-
-	return MyNumRows((*cdaPtr));
     }
+
+    if ((MyQuery(svsock, sbuf, slen) == -1)  &&
+	(!MyReconnect(svsock, h)
+	 ||	 (MyQuery(svsock, sbuf, slen) == -1))) {
+        Safefree(salloc);
+	DO_ERROR(h, JW_ERR_QUERY, svsock);
+	return -2;
+    }
+    Safefree(salloc);
+
+    /** Store the result from the Query */
+#if defined(DBD_MYSQL)
+    if (!(*cdaPtr = (use_mysql_use_result ?
+		     mysql_use_result(svsock) : mysql_store_result(svsock)))) {
+        return mysql_affected_rows(svsock);
+#elif defined(DBD_MSQL)
+    if (!(*cdaPtr = MyStoreResult(svsock))) {
+        return -1;
+#endif
+    }
+
+    return MyNumRows((*cdaPtr));
 }
 
 
@@ -585,6 +879,10 @@ int dbd_st_execute(SV* sth, imp_sth_t* imp_sth) {
      */
     for (i = 0;  i < AV_ATTRIB_LAST;  i++) {
 	if (imp_sth->av_attr[i]) {
+#ifdef DEBUGGING_MEMORY_LEAK
+	    printf("Execute: Decrementing refcnt: old = %d\n",
+		   SvREFCNT(imp_sth->av_attr[i]));
+#endif
 	    SvREFCNT_dec(imp_sth->av_attr[i]);
 	}
 	imp_sth->av_attr[i] = Nullav;
@@ -600,9 +898,9 @@ int dbd_st_execute(SV* sth, imp_sth_t* imp_sth) {
 				     imp_sth->use_mysql_use_result))
 	!= -2) {
 	if (!imp_sth->cda) {
-#xtract Mysql
+#if defined(DBD_MYSQL)
 	    imp_sth->insertid = mysql_insert_id(imp_dbh->svsock);
-#endxtract
+#endif
 	} else {
 	    /** Store the result in the current statement handle */
 	    DBIc_ACTIVE_on(imp_sth);
@@ -662,9 +960,9 @@ AV* dbd_st_fetch(SV* sth, imp_sth_t* imp_sth) {
     int i;
     AV *av;
     row_t cols;
-#xtract Mysql
+#if defined(DBD_MYSQL)
     unsigned int* lengths;
-#endxtract
+#endif
 
     ChopBlanks = DBIc_is(imp_sth, DBIcf_ChopBlanks);
     if (dbis->debug >= 2) {
@@ -678,17 +976,20 @@ AV* dbd_st_fetch(SV* sth, imp_sth_t* imp_sth) {
 
     imp_sth->currow++;
     if (!(cols = MyFetchRow(imp_sth->cda))) {
-#xtract Mysql
+#if defined(DBD_MYSQL)
         if (!mysql_eof(imp_sth->cda)) {
 	    D_imp_dbh_from_sth;
 	    DO_ERROR(sth, JW_ERR_FETCH_ROW, imp_dbh->svsock);
 	}
-#endxtract
+#endif
+	if (!DBIc_COMPAT(imp_sth)) {
+	    dbd_st_finish(sth, imp_sth);
+	}
 	return Nullav;
     }
-#xtract Mysql
+#if defined(DBD_MYSQL)
     lengths = mysql_fetch_lengths(imp_sth->cda);
-#endxtract
+#endif
     av = DBIS->get_fbav(imp_sth);
     num_fields = AvFILL(av)+1;
 
@@ -697,11 +998,11 @@ AV* dbd_st_fetch(SV* sth, imp_sth_t* imp_sth) {
 	SV *sv = AvARRAY(av)[i]; /* Note: we (re)use the SV in the AV	*/
 
 	if (col) {
-#xtract Mysql
+#if defined(DBD_MYSQL)
 	    STRLEN len = lengths[i];
-#xtract Msql
+#elif defined(DBD_MSQL)
 	    STRLEN len = strlen(col);
-#endxtract
+#endif
 	    if (ChopBlanks) {
 		while(len && isspace(col[len-1])) {
 		    --len;
@@ -785,6 +1086,10 @@ void dbd_st_destroy(SV* sth, imp_sth_t* imp_sth) {
      */
     for (i = 0;  i < AV_ATTRIB_LAST;  i++) {
 	if (imp_sth->av_attr[i]) {
+#ifdef DEBUGGING_MEMORY_LEAK
+	    printf("DESTROY: Decrementing refcnt: old = %d\n",
+		   SvREFCNT(imp_sth->av_attr[i]));
+#endif
 	    SvREFCNT_dec(imp_sth->av_attr[i]);
 	}
 	imp_sth->av_attr[i] = Nullav;
@@ -822,11 +1127,11 @@ int dbd_st_STORE_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv, SV* valuesv) {
 		(u_long) sth, key);
     }
 
-#xtract Mysql
+#if defined(DBD_MYSQL)
     if (strEQ(key, "mysql_use_result")) {
         imp_sth->use_mysql_use_result = SvTRUE(valuesv);
     }
-#endxtract
+#endif
 
     if (dbis->debug >= 2) {
         fprintf(DBILOGFP,
@@ -865,19 +1170,19 @@ int dbd_st_STORE_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv, SV* valuesv) {
 #define IS_KEY(A) (((A) & (PRI_KEY_FLAG | UNIQUE_KEY_FLAG | MULTIPLE_KEY_FLAG)) != 0)
 #endif
 #ifndef IS_NUM
-#xtract Mysql
+#if defined(DBD_MYSQL)
 #define IS_NUM(A) ((A) >= (int) FIELD_TYPE_DECIMAL && (A) <= FIELD_TYPE_DATETIME)
-#xtract Msql
+#elif defined(DBD_MSQL)
 #ifdef UINT_TYPE
 #define IS_NUM(A) ((A) == INT_TYPE || (A) == REAL_TYPE || (A) == UINT_TYPE)
 #else
 #define IS_NUM(A) ((A) == INT_TYPE || (A) == REAL_TYPE)
 #endif
-#endxtract
+#endif
 #endif
 
 SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
-    imp_sth_t* imp_sth;
+    D_imp_sth(sth);
     AV *av = Nullav;
     field_t curField;
 
@@ -890,8 +1195,7 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
     /*
      *  Return cached value, if possible
      */
-    } else if (cacheit  &&
-	       (imp_sth = (imp_sth_t*) DBIh_COM(sth))->av_attr[what]) {
+    } else if (cacheit  &&  imp_sth->av_attr[what]) {
 	av = imp_sth->av_attr[what];
 
     /*
@@ -920,6 +1224,56 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
 	      case AV_ATTRIB_TYPE:
 		sv = newSViv((int) curField->type);
 		break;
+	      case AV_ATTRIB_SQL_TYPE:
+		{   int i;
+#ifdef DBD_MYSQL
+		    switch (curField->type) {
+		      case FIELD_TYPE_DECIMAL:     i = SQL_DECIMAL;     break;
+		      case FIELD_TYPE_TINY:        i = SQL_TINYINT;     break;
+		      case FIELD_TYPE_SHORT:       i = SQL_SMALLINT;    break;
+		      case FIELD_TYPE_LONG:        i = SQL_INTEGER;     break;
+		      case FIELD_TYPE_FLOAT:       i = SQL_REAL;        break;
+		      case FIELD_TYPE_DOUBLE:      i = SQL_DOUBLE;      break;
+		      case FIELD_TYPE_TIMESTAMP:   i = SQL_TIMESTAMP;   break;
+		      case FIELD_TYPE_LONGLONG:    i = SQL_BIGINT;      break;
+		      case FIELD_TYPE_INT24:       i = SQL_BIGINT;      break;
+		      case FIELD_TYPE_DATE:        i = SQL_DATE;        break;
+		      case FIELD_TYPE_TIME:        i = SQL_TIME;        break;
+		      case FIELD_TYPE_NEWDATE:     i = SQL_DATE;        break;
+		      case FIELD_TYPE_TINY_BLOB:   i = SQL_VARCHAR;     break;
+		      case FIELD_TYPE_MEDIUM_BLOB: i = SQL_LONGVARCHAR; break;
+		      case FIELD_TYPE_LONG_BLOB:   i = SQL_LONGVARCHAR; break;
+		      case FIELD_TYPE_BLOB:        i = SQL_LONGVARCHAR; break;
+		      case FIELD_TYPE_VAR_STRING:  i = SQL_VARCHAR;     break;
+		      case FIELD_TYPE_STRING:      i = SQL_CHAR;        break;
+		      case FIELD_TYPE_DATETIME:    i = SQL_TIMESTAMP;   break;
+		      case FIELD_TYPE_NULL:        i = SQL_VARCHAR;     break;
+		      case FIELD_TYPE_YEAR:        i = SQL_SMALLINT;    break;
+		      case FIELD_TYPE_ENUM:        i = SQL_CHAR;        break;
+		      case FIELD_TYPE_SET:         i = SQL_CHAR;        break;
+		      default:                     i = SQL_VARCHAR;     break;
+		    }
+#else
+		    switch (curField->type) {
+		        case INT_TYPE:    i = SQL_INTEGER;  break;
+		        case CHAR_TYPE:   i = SQL_VARCHAR;  break;
+		        case REAL_TYPE:   i = SQL_REAL;     break;
+		        case IDENT_TYPE:  i = SQL_VARCHAR;  break;
+		        case NULL_TYPE:   i = SQL_VARCHAR;  break;
+		        case TEXT_TYPE:   i = SQL_VARCHAR;  break;
+		        case DATE_TYPE:   i = SQL_DATE;     break;
+		        case UINT_TYPE:   i = SQL_BIGINT;   break;
+		        case MONEY_TYPE:  i = SQL_VARCHAR;  break;
+		        case TIME_TYPE:   i = SQL_TIME;     break;
+		        case IDX_TYPE:    i = SQL_VARCHAR;  break;
+		        case SYSVAR_TYPE: i = SQL_VARCHAR;  break;
+		        case ANY_TYPE:    i = SQL_VARCHAR;  break;
+		        default:          i = SQL_VARCHAR;  break;
+		    }
+#endif
+		    sv = newSViv(i);
+		}
+		break;
 	      case AV_ATTRIB_IS_PRI_KEY:
 		sv = boolSV(IS_PRI_KEY(curField->flags));
 		break;
@@ -941,7 +1295,7 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
 			int id;
 			const char* name;
 		    } types [] = {
-#xtract Mysql
+#if defined(DBD_MYSQL)
 			{ FIELD_TYPE_BLOB, "blob" },
 			{ FIELD_TYPE_CHAR, "char" },
 			{ FIELD_TYPE_DECIMAL, "decimal" },
@@ -960,7 +1314,7 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
 			{ FIELD_TYPE_TIMESTAMP, "timestamp" },
 			{ FIELD_TYPE_TIME, "time" },
 			{ FIELD_TYPE_VAR_STRING, "varstring" }
-#xtract Msql
+#elif defined(DBD_MSQL)
 			{ INT_TYPE, "int" },
 			{ CHAR_TYPE, "char" },
 			{ REAL_TYPE, "real" },
@@ -986,9 +1340,11 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
 #ifdef SYSVAR_TYPE
 			{ SYSVAR_TYPE, "sys" }
 #endif
-#endxtract
+#endif
 		    };
 		    int i, found = FALSE;
+
+		    sv = &sv_undef;
 		    for (i = 0;  i < sizeof(types) / sizeof(struct db_types);
 			 i++) {
 			if (curField->type == types[i].id) {
@@ -1003,7 +1359,7 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
 		    }
 		}
 	        break;
-#xtract Mysql
+#if defined(DBD_MYSQL)
 	      case AV_ATTRIB_MAX_LENGTH:
 		sv = newSViv((int) curField->max_length);
 		break;
@@ -1013,7 +1369,7 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
 	      case AV_ATTRIB_IS_BLOB:
 		sv = boolSV(IS_BLOB(curField->flags));
 		break;
-#endxtract
+#endif
 	    }
 
 	    av_push(av, sv);
@@ -1026,14 +1382,14 @@ SV* dbd_st_FETCH_internal(SV* sth, int what, result_t res, int cacheit) {
 	if (cacheit) {
 	    imp_sth->av_attr[what] = av;
 	} else {
-	    return sv_2mortal(newRV_noinc((SV*) av));
+	    return sv_2mortal(newRV_noinc((SV*)av));
 	}
     }
 
     if (av == Nullav) {
 	return &sv_undef;
     }
-    return sv_2mortal(newRV((SV*)av));
+    return sv_2mortal(newRV_inc((SV*)av));
 }
 
 
@@ -1067,6 +1423,14 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
     }
 
     switch (*key) {
+#ifdef DBD_MYSQL
+      case 'a':
+	if (strEQ(key, "affected_rows")) {
+	    D_imp_dbh_from_sth;
+	    retsv = sv_2mortal(newSViv((IV)mysql_affected_rows(imp_dbh->svsock)));
+	}
+	break;
+#endif
       case 'I':
 	/*
 	 *  Deprecated, use lower case versions.
@@ -1075,12 +1439,12 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_PRI_KEY);
 	} else if (strEQ(key, "IS_NOT_NULL")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_NOT_NULL);
-#xtract Mysql
+#if defined(DBD_MYSQL)
 	} else if (strEQ(key, "IS_KEY")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_KEY);
 	} else if (strEQ(key, "IS_BLOB")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_BLOB);
-#endxtract
+#endif
 	} else if (strEQ(key, "IS_NUM")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_NUM);
 	}
@@ -1093,7 +1457,7 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_LENGTH);
 	}
 	break;
-#xtract Mysql
+#if defined(DBD_MYSQL)
       case 'M':
 	/*
 	 *  Deprecated, use max_length
@@ -1102,7 +1466,7 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_MAX_LENGTH);
 	}
 	break;
-#endxtract
+#endif
       case 'N':
 	if (strEQ(key, "NAME")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_NAME);
@@ -1123,21 +1487,23 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
 	}
 	break;
       case 'T':
-	/*
-	 *  Deprecated, use lower case versions.
-	 */
-	if (strEQ(key, "TABLE")) {
-	    retsv = ST_FETCH_AV(AV_ATTRIB_TABLE);
-	} else if (strEQ(key, "TYPE")) {
-	    retsv = ST_FETCH_AV(AV_ATTRIB_TYPE);
+	if (strEQ(key, "TYPE")) {
+	    retsv = ST_FETCH_AV(AV_ATTRIB_SQL_TYPE);
 	}
+	/*
+	 *  Deprecated, use lower case version.
+	 */
+	else if (strEQ(key, "TABLE")) {
+	    retsv = ST_FETCH_AV(AV_ATTRIB_TABLE);
+	}
+	break;
       case 'f':
 	if (strEQ(key, "format_max_size")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_LENGTH);
-#xtract Mysql
+#if defined(DBD_MYSQL)
 	} else if (strEQ(key, "format_default_size")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_MAX_LENGTH);
-#endxtract
+#endif
 	} else if (strEQ(key, "format_right_justify")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_NUM);
 	} else if (strEQ(key, "format_type_name")) {
@@ -1151,12 +1517,12 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_PRI_KEY);
 	} else if (strEQ(key, "is_not_null")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_NOT_NULL);
-#xtract Mysql
+#if defined(DBD_MYSQL)
 	} else if (strEQ(key, "is_key")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_KEY);
 	} else if (strEQ(key, "is_blob")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_BLOB);
-#endxtract
+#endif
 	} else if (strEQ(key, "is_num")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_IS_NUM);
 	}
@@ -1166,19 +1532,22 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_LENGTH);
 	}
 	break;
-#xtract Mysql
       case 'm':
+#if defined(DBD_MYSQL)
 	if (strEQ(key, "max_length")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_MAX_LENGTH);
 	} else if (strEQ(key, "mysql_use_result")) {
 	    retsv = boolSV(imp_sth->use_mysql_use_result);
+	} else if (strEQ(key, "mysql_type")) {
+	    retsv = ST_FETCH_AV(AV_ATTRIB_TYPE);
 	}
+#else
+	if (strEQ(key, "msql_type")) {
+	    retsv = ST_FETCH_AV(AV_ATTRIB_TYPE);
+	}
+#endif
 	break;
-#endxtract
       case 'r':
-	/*
-	 * Deprecated, use 'result'
-	 */
 	if (strEQ(key, "result")) {
 	    retsv = sv_2mortal(newSViv((IV) imp_sth->cda));
 	}
@@ -1186,9 +1555,8 @@ SV* dbd_st_FETCH_attrib(SV* sth, imp_sth_t* imp_sth, SV* keysv) {
       case 't':
 	if (strEQ(key, "table")) {
 	    retsv = ST_FETCH_AV(AV_ATTRIB_TABLE);
-	} else if (strEQ(key, "type")) {
-	    retsv = ST_FETCH_AV(AV_ATTRIB_TYPE);
 	}
+	break;
     }
 
     if (dbis->debug >= 2) {
@@ -1286,3 +1654,39 @@ int dbd_bind_ph (SV *sth, imp_sth_t *imp_sth, SV *param, SV *value,
 
     return BindParam(&imp_sth->params[paramNum - 1], value, sql_type);
 }
+
+
+#if defined(DBD_MYSQL)
+
+/***************************************************************************
+ *
+ *  Name:    MysqlReconnect
+ *
+ *  Purpose: If the server has disconnected, try to reconnect.
+ *
+ *  Input:   h - database or statement handle
+ *
+ *  Returns: TRUE for success, FALSE otherwise
+ *
+ **************************************************************************/
+
+int MysqlReconnect(SV* h) {
+    D_imp_xxh(h);
+    imp_dbh_t* imp_dbh;
+
+    if (DBIc_TYPE(imp_xxh) == DBIt_ST) {
+        imp_dbh = (imp_dbh_t*) DBIc_PARENT_COM(imp_xxh);
+	h = DBIc_PARENT_H(imp_xxh);
+    } else {
+        imp_dbh = (imp_dbh_t*) imp_xxh;
+    }
+
+    if (!_MyLogin(imp_dbh)) {
+	DO_ERROR(h, MyErrno(imp_dbh->svsock, JW_ERR_CONNECT),
+		 imp_dbh->svsock);
+	return FALSE;
+    }
+    return TRUE;
+}
+
+#endif
